"""Top-level imports"""

from . import _version
from .enderchest import EnderChest
from .instance import InstanceSpec
from .shulker_box import ShulkerBox

__version__ = _version.get_versions()["version"]


__all__ = [
    "EnderChest",
    "InstanceSpec",
    "ShulkerBox",
]



"""Subpackage containing all files and templates used for testing"""

from importlib.resources import files

__all__ = [
    "CLIENT_OPTIONS",
    "ENDERCHEST_CONFIG",
    "LSTAT_CACHE",
    "INSTGROUPS",
    "LAUNCHER_PROFILES",
    "VERSION_MANIFEST",
]

_here = files(__package__)

CLIENT_OPTIONS = _here / "options.txt"

ENDERCHEST_CONFIG = _here / "enderchest.cfg"

LSTAT_CACHE = _here / "lstat_cache.json"

INSTGROUPS = _here / "instgroups.json"

LAUNCHER_PROFILES = _here / "launcher_profiles.json"

VERSION_MANIFEST = _here / "version_manifest_v2.json"



"""Tests for EnderChest / Shulker Box state resolving functionality"""

import logging
import re
from pathlib import Path

import pytest

from enderchest import craft
from enderchest import filesystem as fs
from enderchest import inventory
from enderchest.test import utils


class TestLoadEnderChestInstances:
    def test_bad_folder_just_returns_empty(self, tmp_path, caplog):
        assert inventory.load_ender_chest_instances(tmp_path) == []


class TestListShulkerBoxes:
    @pytest.fixture(autouse=True)
    def populate_shulker_boxes(self, minecraft_root):
        utils.pre_populate_enderchest(
            minecraft_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS
        )

        # also write a bad shulker box
        bad_ini = minecraft_root / "EnderChest" / "not_ini" / fs.SHULKER_BOX_CONFIG_NAME
        bad_ini.parent.mkdir()
        bad_ini.write_text("is_this_valid_ini=no")

    def test_bad_folder_just_returns_empty(self, tmp_path, caplog):
        assert inventory.load_shulker_boxes(tmp_path) == []

    def test_warn_for_a_chest_without_boxes(self, tmp_path, caplog):
        root = tmp_path / "nowhere"
        root.mkdir()
        craft.craft_ender_chest(root, remotes=(), overwrite=True)
        _ = inventory.load_shulker_boxes(root)

        assert "There are no shulker boxes" in "\n".join(
            (
                record.msg
                for record in caplog.records
                if record.levelno == logging.WARNING
            )
        )

    def test_list_shulker_box_reports_the_boxes_in_order(self, minecraft_root, caplog):
        _ = inventory.load_shulker_boxes(minecraft_root)
        assert (
            """0. global
  1. 1.19
  2. vanilla
  3. optifine"""
            in caplog.records[-1].msg % caplog.records[-1].args
        )

    def test_list_shulker_box_doesnt_choke_on_relative_root(
        self, minecraft_root, caplog, monkeypatch
    ):
        monkeypatch.chdir(minecraft_root.parent)
        _ = inventory.load_shulker_boxes(Path(minecraft_root.name))
        assert (
            """0. global
  1. 1.19
  2. vanilla
  3. optifine"""
            in caplog.records[-1].msg % caplog.records[-1].args
        )

    def test_list_shulker_box_warns_if_theres_a_bad_box(self, minecraft_root, caplog):
        _ = inventory.load_shulker_boxes(minecraft_root)

        warnings = [
            record for record in caplog.records if record.levelname == "WARNING"
        ]

        assert re.search(
            rf"Could not parse(.*)not_ini(.*){fs.SHULKER_BOX_CONFIG_NAME}",
            warnings[-1].msg % warnings[-1].args,
        )


class TestLoadBoxInstanceMatches:
    @pytest.fixture(autouse=True)
    def populate_shulker_boxes(self, minecraft_root, home):
        utils.pre_populate_enderchest(
            minecraft_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS
        )

    def test_bad_folder_just_returns_empty(self, tmp_path, caplog):
        assert (
            inventory.get_shulker_boxes_matching_instance(tmp_path, "outstance") == []
        )

    @pytest.mark.parametrize(
        "instance_name",
        [mc.name for mc in utils.TESTING_INSTANCES] + ["unown"],
    )
    def test_loading_boxes_that_match_instance(self, minecraft_root, instance_name):
        # because TESTING_SHULKER_INSTANCE_MATCHES uses the path
        instance_name_lookup = {
            "official": "~",
            "Chest Boat": "chest-boat",
            "Drowned": "drowned",
        }
        box_lookup = {
            box.name: box
            for box in inventory.load_shulker_boxes(
                minecraft_root, log_level=logging.DEBUG
            )
        }

        expected = []
        for box_name, mc_name, should_match in utils.TESTING_SHULKER_INSTANCE_MATCHES:
            if mc_name == instance_name_lookup.get(instance_name, instance_name):
                if should_match:
                    expected.append(box_lookup[box_name])

        assert (
            inventory.get_shulker_boxes_matching_instance(minecraft_root, instance_name)
            == expected
        )

    @pytest.mark.parametrize(
        "shulker_box_name",
        list({box for _, box, _ in utils.TESTING_SHULKER_INSTANCE_MATCHES}) + ["unown"],
    )
    def test_loading_instances_that_match_boxes(self, minecraft_root, shulker_box_name):
        instance_lookup = {
            mc.name: mc
            for mc in inventory.load_ender_chest_instances(
                minecraft_root, log_level=logging.DEBUG
            )
        }
        # because TESTING_SHULKER_INSTANCE_MATCHES uses the path
        instance_lookup["~"] = instance_lookup["official"]
        instance_lookup["chest-boat"] = instance_lookup["Chest Boat"]

        expected = []
        for (
            box_name,
            instance_name,
            should_match,
        ) in utils.TESTING_SHULKER_INSTANCE_MATCHES:
            if box_name == shulker_box_name:
                if should_match:
                    expected.append(instance_lookup[instance_name])

        assert (
            inventory.get_instances_matching_shulker_box(
                minecraft_root, shulker_box_name
            )
            == expected
        )



"""Additional pytest CLI options. h/t https://stackoverflow.com/a/52458082"""


def pytest_addoption(parser):
    parser.addoption(
        "--use-local-ssh",
        action="store_true",
        default=False,
        help=(
            "By default, SFTP tests are run against a mock SSH server."
            " If your system happens to be set up for passwordless sshing"
            " into localhost, you can choose to run your tests against your"
            " actual SSH server by using this flag."
        ),
    )



"""Tests of the sync-helper utilities"""

import importlib
import logging
import os
import shutil
import time
from pathlib import Path
from urllib.parse import urlparse

import pytest

from enderchest import sync
from enderchest.sync import file
from enderchest.sync import utils as sync_utils
from enderchest.sync.utils import Operation as Op


class TestDetermineAvailableProtocols:
    def test_available_protocols_only_include_resolvable_modules(self, monkeypatch):
        monkeypatch.setattr(sync, "PROTOCOLS", ("file", "subspace"))
        assert sync._determine_available_protocols() == ("file",)


class TestPathFromURI:
    def test_roundtrip(self, tmpdir):
        original_path = Path(tmpdir) / "this has a space in it" / "(="
        assert (
            sync_utils.abspath_from_uri(urlparse(original_path.as_uri()))
            == original_path
        )


class TestURIToSSH:
    def test_simple_parse(self):
        address = sync_utils.uri_to_ssh(
            urlparse("rsync://openbagtwo@couchgaming:22/home/openbagtwo/minecraft")
        )
        assert address == "openbagtwo@couchgaming:22:/home/openbagtwo/minecraft"

    def test_no_username_parse(self):
        address = sync_utils.uri_to_ssh(
            urlparse("rsync://steamdeck/home/openbagtwo/minecraft")
        )
        assert address == "steamdeck:/home/openbagtwo/minecraft"

    def test_no_netloc_parse(self):
        address = sync_utils.uri_to_ssh(urlparse("rsync:///mnt/external/minecraft-bkp"))
        assert address == "localhost:/mnt/external/minecraft-bkp"

    def test_no_hostname_parse(self):
        """Can't believe this is a valid URI"""
        address = sync_utils.uri_to_ssh(urlparse("rsync://nugget@/home/nugget/"))
        assert address == "nugget@localhost:/home/nugget/"


class TestRenderRemote:
    @pytest.mark.parametrize("alias", ("matching", "not-matching"))
    def test_render_omits_alias_only_if_it_matches_hostname(self, alias):
        uri = "prtcl://youser@matching:101/some/directory"
        rendered = sync_utils.render_remote(alias, urlparse(uri))
        if alias == "matching":
            assert rendered == uri
        else:
            assert rendered == uri + " (not-matching)"


class TestIsIdentical:
    def test_copied_files_are_identical(self, tmp_path):
        one = tmp_path / "one"
        one.write_text("I'm the original\n")
        two = tmp_path / "two"
        shutil.copy2(one, two)
        assert sync_utils.is_identical(one.stat(), two.stat())

    def test_modified_files_are_not_identical(self, tmp_path):
        one = tmp_path / "one"
        one.write_text("I'm the original")
        two = tmp_path / "two"
        shutil.copy2(one, two)
        with two.open("a") as f:
            f.write("\n")
        assert not sync_utils.is_identical(one.stat(), two.stat())

    @pytest.mark.parametrize("target_type", ("file", "folder"))
    def test_symlink_is_not_identical_to_its_target(self, tmp_path, target_type):
        target = tmp_path / "bulleye"
        if target_type == "file":
            target.touch()
        else:
            target.mkdir()

        linky = tmp_path / "linky"
        linky.symlink_to(target, target_is_directory=target_type == "folder")

        assert not sync_utils.is_identical(target.lstat(), linky.lstat())

    def test_checks_modified_time(self, tmp_path):
        one = tmp_path / "one"
        one.write_text("I'm the original")
        two = tmp_path / "two"
        shutil.copy(one, two)
        time.sleep(1)  # ugh
        two.write_text("I'm the original")

        # meta-test that the file contents are identical
        assert one.read_bytes() == two.read_bytes()

        assert not sync_utils.is_identical(one.stat(), two.stat())

    def test_does_not_check_modified_time_of_directories(self, tmp_path):
        one = tmp_path / "one"
        one.mkdir(parents=True)
        two = tmp_path / "two"
        time.sleep(1)  # ugh
        two.mkdir()
        assert sync_utils.is_identical(one.stat(), two.stat())

    @pytest.mark.xfail(reason="Method does not compute hashes")
    def test_two_files_of_the_same_size_are_not_identical(self, tmp_path):
        one = tmp_path / "one"
        one.write_text("I'm the original")
        two = tmp_path / "two"
        two.write_text("I'm the octagonl")

        assert not sync_utils.is_identical(one.stat(), two.stat())


@pytest.fixture
def file_system(tmp_path):
    source_tree = tmp_path / "oak"
    (source_tree / "branch").mkdir(parents=True)
    (source_tree / "branch" / "twig").mkdir()
    (source_tree / "branch" / "leaf").write_text("green")
    (source_tree / "branch" / "acorn").touch()
    (source_tree / "branch2").mkdir()
    (source_tree / "branch2" / "acorn").touch()

    dst_tree = tmp_path / "birch"
    (dst_tree / "branch").mkdir(parents=True)
    (dst_tree / "branch" / "twig").mkdir()
    (dst_tree / "branch" / "leaf").write_text("yellow")
    shutil.copy2((source_tree / "branch" / "acorn"), (dst_tree / "branch" / "acorn"))
    (dst_tree / "beehive").touch()
    (dst_tree / "root").mkdir()
    (dst_tree / "root" / "cicada").touch()

    yield tmp_path


@pytest.fixture
def diff(file_system):
    yield sync_utils.diff(
        file.get_contents(file_system / "oak"),
        file.get_contents(file_system / "birch"),
    )


class TestDiff:
    def test_diff_omits_identical_contents(self, diff):
        assert {
            Path("branch") / "acorn",
            Path("branch") / "twig",
            Path("branch"),
        }.intersection((f for f, *_ in diff)) == set()

    def test_diff_captures_updates(self, diff):
        assert [Path("branch") / "leaf"] == [f for f, _, op in diff if op == Op.REPLACE]

    def test_diff_captures_creation_of_files_and_folders(self, diff):
        assert [
            Path("branch2"),
            Path("branch2") / "acorn",
        ] == [f for f, _, op in diff if op == Op.CREATE]

    def test_diff_captures_deletion_of_files_and_folders(self, diff):
        assert [
            Path("root") / "cicada",
            Path("beehive"),
            Path("root"),
        ] == [f for f, _, op in diff if op == Op.DELETE]


class TestGenerateSyncReport:
    def test_generate_sync_report(self, diff, caplog):
        expected = [
            "Deleting beehive",
            """Within branch...
  - Creating 0 files
  - Replacing 1 file
  - Deleting 0 files""",
            "Creating branch2",
            "Deleting root",
        ]

        sync_utils.generate_sync_report(diff)

        info_logs = [
            (record.msg % record.args).strip()
            for record in caplog.records
            if record.levelno == logging.INFO
        ]

        assert info_logs == expected


class TestFileIgnorePatternBuilder:
    def test_simple_match(self):
        assert file.ignore_patterns("hello")(
            "greetings", ("bonjour", "hello", "hellooooo")
        ) == {"hello"}

    def test_wildcard(self):
        assert file.ignore_patterns("hel*")(
            "responses", ("como sa va", "hello", "hell no", "help")
        ) == {"hello", "hell no", "help"}

    def test_multi_pattern_match(self):
        assert file.ignore_patterns("hel*", "bye")(
            "responses", ("hello", "goodbye", "hellooo", "bye")
        ) == {"hello", "hellooo", "bye"}

    def test_full_path_check(self):
        ignore = file.ignore_patterns(os.path.join("root", "branch"))
        assert (
            ignore("root", ("branch", "trunk")),
            ignore("trunk", ("branch", "leaf")),
        ) == ({"branch"}, set())

    def test_match_is_performed_on_the_end(self):
        ignore = file.ignore_patterns(os.path.join("root", "branch"), "leaf")
        assert (
            ignore(os.path.join("tree", "root"), ("branch", "trunk")),
            ignore("wind", ("leaf", "blows")),
        ) == ({"branch"}, {"leaf"})


class TestFileClean:
    def test_clean_cleans_recursively_while_respecting_ignore(self, tmp_path):
        root = tmp_path / "root"

        (root / "empty dir").mkdir(parents=True)
        (root / "will be empty").mkdir()
        (root / "will be empty" / "filey").touch()
        (root / "will not be empty").mkdir()
        (root / "will not be empty" / "file-o bread").touch()
        (root / "will not be empty" / "important.txt").write_text("leave me be\n")
        (root / "dont go inside").mkdir()
        (root / "dont go inside" / "keep me safe").touch()
        (root / "tricky linky").symlink_to(root / "will not be empty")
        (root / "trickier linkier").symlink_to(root / "dont go inside")

        file.clean(root, ignore=file.ignore_patterns("*.txt", "dont*"), dry_run=False)

        assert set(root.rglob("**/*")) == {
            root / "will not be empty",
            root / "will not be empty" / "important.txt",
            root / "dont go inside",
            root / "dont go inside" / "keep me safe",
        }



"""Tests around file transfer functionality"""

import itertools
import json
import logging
import os
import shutil
import socket
import subprocess
import sys
import time
from importlib.resources import as_file
from pathlib import Path
from urllib.parse import ParseResult, urlparse

import pytest

from enderchest import craft
from enderchest import filesystem as fs
from enderchest import gather, inventory, place
from enderchest import remote as r
from enderchest import sync
from enderchest.sync import utils as sync_utils

from . import mock_paramiko, utils
from .testing_files import LSTAT_CACHE


@pytest.fixture(autouse=True)
def force_hostname_to_be_localhost(monkeypatch):
    if sys.platform == "darwin":  # so far this is only needed for macOS
        monkeypatch.setattr(socket, "gethostname", lambda: "localhost")


class TestFileSync:
    protocol = "file"

    @pytest.fixture(autouse=True)
    def check_output_suppression(self, capfd):
        # teardown that helps me figure out where I need to add a -q flag
        yield
        assert capfd.readouterr().out == ""

    @pytest.fixture
    def remote(self, tmp_path, minecraft_root, home):
        utils.pre_populate_enderchest(
            minecraft_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS
        )

        local = inventory.load_ender_chest(minecraft_root)
        local.sync_confirm_wait = False
        local.write_to_cfg(fs.ender_chest_config(minecraft_root))

        another_root = tmp_path / "not so remote"

        (another_root / "EnderChest").mkdir(parents=True)

        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS[1:]:
            shutil.copytree(
                minecraft_root / "EnderChest" / folder_name,
                another_root / "EnderChest" / folder_name,
                symlinks=True,
            )

        (minecraft_root / "safe_keeping").mkdir()
        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS[:-2]:
            shutil.copytree(
                minecraft_root / "EnderChest" / folder_name,
                minecraft_root / "safe_keeping" / folder_name,
                symlinks=True,
            )
        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS[-2:]:
            shutil.move(
                minecraft_root / "EnderChest" / folder_name,
                minecraft_root / "safe_keeping" / folder_name,
            )

        utils.pre_populate_enderchest(
            another_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS[1:]
        )

        not_so_remote = inventory.load_ender_chest(another_root)
        not_so_remote.name = "closer than you think"
        not_so_remote._uri = not_so_remote._uri._replace(scheme=self.protocol)
        not_so_remote.register_remote(local._uri)
        not_so_remote.register_remote("ipoac://yoursoul@birdhouse/minecraft")
        not_so_remote.do_not_sync.append("EnderChest/*.local")
        not_so_remote.write_to_cfg(another_root / "EnderChest" / "enderchest.cfg")

        (another_root / "chest monster" / "worlds" / "olam").mkdir(parents=True)
        (another_root / "chest monster" / "worlds" / "olam" / "level.dat").write_text(
            "dootdootdoot"
        )

        (another_root / "EnderChest" / "1.19" / "saves" / "olam").unlink()
        (another_root / "EnderChest" / "1.19" / "saves" / "olam").symlink_to(
            another_root / "chest monster" / "worlds" / "olam", target_is_directory=True
        )

        (another_root / "EnderChest" / "1.19" / ".bobby").mkdir(exist_ok=True)
        (another_root / "EnderChest" / "1.19" / ".bobby" / "chunk").write_text(
            "chunky\n"
        )

        for root in (minecraft_root, another_root):
            (root / "EnderChest" / "vanilla" / "conflict").mkdir()

        (
            minecraft_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).write_text("sparkle")
        (
            another_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).write_text("lab-grown!")

        (another_root / "EnderChest" / "optifine" / "mods").mkdir(
            parents=True, exist_ok=True
        )
        (another_root / "EnderChest" / "optifine" / "mods" / "optifine.jar").write_text(
            "it's okay"
        )

        (minecraft_root / "EnderChest" / "1.19" / "resourcepacks").mkdir()
        (another_root / "EnderChest" / "1.19" / "resourcepacks").mkdir()
        (
            minecraft_root
            / "EnderChest"
            / "1.19"
            / "resourcepacks"
            / "TEAVSRP_lite.zip"
        ).write_text("I am haggling you\n")
        shutil.copy2(
            (
                minecraft_root
                / "EnderChest"
                / "1.19"
                / "resourcepacks"
                / "TEAVSRP_lite.zip"
            ),
            (
                another_root
                / "EnderChest"
                / "1.19"
                / "resourcepacks"
                / "TEAVSRP_lite.zip"
            ),
        )

        (another_root / "EnderChest" / "really.local").mkdir()
        (another_root / "EnderChest" / "really.local" / "all_mine.txt").write_text(
            "DO NOT SYNC!!!\n"
        )

        yield not_so_remote._uri

        assert (
            another_root / "chest monster" / "worlds" / "olam" / "level.dat"
        ).read_text() == "dootdootdoot"

        assert (
            another_root / "EnderChest" / "really.local" / "all_mine.txt"
        ).read_text("UTF-8") == "DO NOT SYNC!!!\n"

        assert not (
            minecraft_root / "EnderChest" / "really.local" / "all_mine.txt"
        ).exists()

        # now put everything local back the way it was supposed to be before
        # the conftest teardown freaks out

        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS:
            shutil.move(
                minecraft_root / "safe_keeping" / folder_name,
                minecraft_root / "EnderChest" / folder_name,
            )

        # and then I apparently still need to redo some symlinks?
        for path, target in (
            (
                minecraft_root / "EnderChest" / "global" / "crash-reports",
                minecraft_root / "crash-reports",
            ),
            (
                minecraft_root / "EnderChest" / "global" / "resourcepacks",
                minecraft_root / "resourcepacks",
            ),
            (
                minecraft_root / "EnderChest" / "global" / "screenshots",
                home / "Pictures" / "Screenshots",
            ),
            (
                minecraft_root / "EnderChest" / "global" / "saves" / "test",
                minecraft_root / "worlds" / "testbench",
            ),
            (
                minecraft_root / "EnderChest" / "1.19" / "saves" / "olam",
                minecraft_root / "worlds" / "olam",
            ),
        ):
            path.unlink(missing_ok=True)
            path.symlink_to(target, target_is_directory=target.is_dir())

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_create_from_remote_chest(
        self, remote, minecraft_root, monkeypatch, root_type, capfd
    ):
        if root_type == "absolute":
            root = minecraft_root
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)

        craft.craft_ender_chest(root, copy_from=remote.geturl(), overwrite=True)

        assert inventory.load_ender_chest_remotes(minecraft_root) == [
            (remote, "closer than you think"),
            (urlparse("ipoac://yoursoul@birdhouse/minecraft"), "birdhouse"),
        ]

        # test that the enderchest spec-fetch is quiet by default
        assert capfd.readouterr().out == ""

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_open_grabs_files_from_upstream(
        self, minecraft_root, remote, monkeypatch, root_type
    ):
        if root_type == "absolute":
            root = minecraft_root
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)

        gather.update_ender_chest(root, remotes=(remote,))
        r.sync_with_remotes(root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "optifine" / "mods" / "optifine.jar"
        ).read_text() == "it's okay"
        assert not (
            minecraft_root / "EnderChest" / "optifine" / "mods" / "optifine.jar"
        ).is_symlink()

    def test_open_dry_run_does_nothing(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", dry_run=True)
        assert not (
            minecraft_root / "EnderChest" / "optifine" / "mods" / "optifine.jar"
        ).exists()

    def test_open_overwrites_local_files(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).read_text() == "lab-grown!"
        assert not (
            minecraft_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).is_symlink()

    def test_open_copies_over_symlinks(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "1.19" / "saves" / "olam"
        ).resolve() != (minecraft_root / "worlds" / "olam")

    @pytest.mark.parametrize("delete", (True, False), ids=("default", "do-not-delete"))
    def test_open_respects_delete_kwarg_when_processing_deletions_from_upstream(
        self, minecraft_root, remote, delete
    ):
        sync_kwargs = {"verbosity": -1}
        if not delete:
            sync_kwargs["delete"] = False
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", **sync_kwargs)
        assert not (minecraft_root / "EnderChest" / "global").exists() == delete

    def test_open_does_not_overwrite_enderchest_by_default(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        original_config = fs.ender_chest_config(minecraft_root).read_text()
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert original_config == fs.ender_chest_config(minecraft_root).read_text()

    def test_open_does_not_touch_top_level_dot_folders_by_default(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / ".git" / "log"
        ).read_text() == "i committed some stuff\n"

    def test_open_will_sync_dot_folders_within_a_shulker_box(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "1.19" / ".bobby" / "chunk"
        ).read_text() == "chunky\n"

    @pytest.mark.parametrize(
        "dry_run, place_after",
        (
            (False, False),
            (False, True),
            # no point in testing dry-run-False
            (True, True),
        ),
        ids=("False", "True", "dry-run"),
    )
    def test_place_after_open(self, minecraft_root, remote, dry_run, place_after):
        test_path = (
            minecraft_root
            / "instances"
            / "axolotl"
            / ".minecraft"
            / "conflict"
            / "diamond.png"
        )

        # meta-test
        place.place_ender_chest(minecraft_root)
        assert test_path.read_text("utf-8") == "sparkle"

        enderchest = inventory.load_ender_chest(minecraft_root)
        enderchest.register_remote(remote, alias="not so remote")
        enderchest.place_after_open = place_after
        enderchest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        r.sync_with_remotes(minecraft_root, "pull", dry_run=dry_run, verbosity=-1)
        assert (
            test_path.read_text("utf-8") == "lab-grown!"
            if place_after and not dry_run
            else "sparkle"
        )

    @pytest.mark.parametrize("fail_type", ("no_remotes", "bad_remotes"))
    def test_does_not_place_after_failed_open(
        self, minecraft_root, remote, fail_type, caplog
    ):
        # TODO: move this into TestFileSyncOnly as there is no value in testing
        #       this for multiple protocols

        test_path = (
            minecraft_root
            / "instances"
            / "axolotl"
            / ".minecraft"
            / "conflict"
            / "diamond.png"
        )
        enderchest = inventory.load_ender_chest(minecraft_root)
        if fail_type == "bad_remotes":
            enderchest.register_remote("file://i/do/not/exist", "does not exist")
        enderchest.place_after_open = True
        enderchest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)

        failure_messages = {
            "no_remotes": "EnderChest has no remotes. Aborting.",
            "bad_remotes": "Could not sync with any remote EnderChests",
        }

        assert [  # meta-test
            record.msg for record in caplog.records if record.levelno == logging.ERROR
        ] == [failure_messages[fail_type]]

        assert not test_path.exists()

    def test_does_not_place_after_dry_run(self, minecraft_root, remote, caplog):
        # TODO: move this into TestFileSyncOnly as there is no value in testing
        #       this for multiple protocols

        test_path = (
            minecraft_root
            / "instances"
            / "axolotl"
            / ".minecraft"
            / "conflict"
            / "diamond.png"
        )
        enderchest = inventory.load_ender_chest(minecraft_root)
        enderchest.register_remote(remote, alias="not so remote")
        enderchest.place_after_open = True
        enderchest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1, dry_run=True)

        assert not test_path.exists()

    @pytest.mark.parametrize("operation", ("pull", "push"))
    def test_timeout_argument_doesnt_obviously_break_(
        self, minecraft_root, remote, operation
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, operation, verbosity=-1, timeout=15)

    @pytest.mark.parametrize("operation", ("pull", "push"))
    def test_identical_objects_are_not_synced(
        self, minecraft_root, remote, caplog, operation
    ):
        caplog.set_level(logging.DEBUG)
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, operation, verbosity=-1)
        debug_logs = [
            record.msg % record.args
            for record in caplog.records
            if record.levelno == logging.DEBUG
        ]
        assert debug_logs  # meta-test
        assert not [message for message in debug_logs if "TEAVSRP_lite.zip" in message]

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_close_overwrites_with_changes_from_local(
        self, minecraft_root, remote, monkeypatch, root_type
    ):
        if root_type == "absolute":
            root = minecraft_root
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)

        gather.update_ender_chest(root, remotes=(remote,))
        r.sync_with_remotes(root, "push", verbosity=-1)
        assert (
            fs.shulker_box_root(sync.abspath_from_uri(remote), "vanilla")
            / "conflict"
            / "diamond.png"
        ).read_text() == "sparkle"

    @pytest.mark.parametrize("operation", ("pull", "push"))
    def test_objects_are_identical_after_sync(
        self, minecraft_root, remote, caplog, operation
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        time.sleep(1)  # ugh
        r.sync_with_remotes(minecraft_root, operation, verbosity=-1)
        assert sync_utils.is_identical(
            (
                fs.shulker_box_root(minecraft_root, "vanilla")
                / "conflict"
                / "diamond.png"
            ).stat(),
            (
                fs.shulker_box_root(sync.abspath_from_uri(remote), "vanilla")
                / "conflict"
                / "diamond.png"
            ).stat(),
        )

    def test_close_dry_run_does_nothing(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "push", dry_run=True)
        assert (
            fs.shulker_box_root(sync.abspath_from_uri(remote), "vanilla")
            / "conflict"
            / "diamond.png"
        ).read_text() == "lab-grown!"

    @pytest.mark.parametrize("delete", (True, False), ids=("default", "do-not-delete"))
    def test_close_respects_delete_kwarg_remote_copies_when_locals_are_deleted(
        self,
        minecraft_root,
        remote,
        delete,
    ):
        sync_kwargs = {"verbosity": -1}
        if not delete:
            sync_kwargs["delete"] = False
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "push", **sync_kwargs)
        assert (
            not (
                fs.ender_chest_folder(sync.abspath_from_uri(remote)) / "optifine"
            ).exists()
            == delete
        )

    def test_close_does_not_touch_top_level_dot_folders_by_default(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        assert not (
            fs.ender_chest_folder(sync.abspath_from_uri(remote)) / ".git"
        ).exists()

    def test_chest_obeys_its_own_ignore_list(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))

        chest = inventory.load_ender_chest(minecraft_root)
        chest.do_not_sync.append("super_secret.txt")
        chest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        (
            fs.shulker_box_root(minecraft_root, "vanilla") / "super_secret.txt"
        ).write_text("I'm not sharing!!!")

        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        assert not (
            fs.shulker_box_root(sync.abspath_from_uri(remote), "vanilla")
            / "super_secret.txt"
        ).exists()

        # and then just to make sure
        assert (
            fs.shulker_box_root(minecraft_root, "vanilla") / "super_secret.txt"
        ).read_text("UTF-8") == "I'm not sharing!!!"

    def test_syncing_a_default_exclusion_requires_changes_on_both_sides(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))

        chest = inventory.load_ender_chest(minecraft_root)
        chest.do_not_sync = ["EnderChest/enderchest.cfg"]
        chest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        remote_chest = r.load_remote_ender_chest(remote)
        remote_chest.do_not_sync.remove("EnderChest/.*")
        remote_chest.write_to_cfg(fs.ender_chest_config(sync.abspath_from_uri(remote)))

        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        assert (
            sync.abspath_from_uri(remote) / "EnderChest" / ".git" / "log"
        ).read_text() == "i committed some stuff\n"

    def test_open_stops_at_first_successful_sync(self, minecraft_root, remote, caplog):
        gather.update_ender_chest(
            minecraft_root, remotes=(remote, "prayer://unreachable")
        )
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        warnings = [
            record.msg for record in caplog.records if record.levelname == "WARNING"
        ]

        assert len(warnings) == 0

    def test_close_will_attempt_to_push_to_all(self, minecraft_root, remote, caplog):
        gather.update_ender_chest(
            minecraft_root, remotes=(remote, "prayer://unreachable")
        )
        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        warnings = [
            record.getMessage()
            for record in caplog.records
            if record.levelname == "WARNING"
        ]

        assert len(warnings) == 1
        assert "Could not sync changes with prayer://unreachable" in warnings[0]

    def test_never_places_after_close(self, minecraft_root, remote):
        test_path = (
            minecraft_root
            / "instances"
            / "bee"
            / ".minecraft"
            / "mods"
            / "optifine.jar"
        )

        enderchest = inventory.load_ender_chest(minecraft_root)
        enderchest.register_remote(remote, alias="not so remote")
        enderchest.place_after_open = True
        enderchest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        assert not test_path.exists()

    @pytest.mark.parametrize(
        "operation, mode",
        list(itertools.product(("pull", "push"), ("immediate", "dry_run_first"))),
    )
    def test_sync_respects_exclude(
        self, minecraft_root, remote, caplog, operation, mode
    ):
        sync_confirm_wait = 1 if mode == "dry_run_first" else 0
        gather.update_ender_chest(minecraft_root, remotes=(remote,))

        r.sync_with_remotes(
            minecraft_root,
            operation,
            verbosity=-1,
            sync_confirm_wait=sync_confirm_wait,
            exclude="*/diamond.png",
        )
        assert (
            fs.shulker_box_root(minecraft_root, "vanilla") / "conflict" / "diamond.png"
        ).read_text() != (
            fs.shulker_box_root(sync.abspath_from_uri(remote), "vanilla")
            / "conflict"
            / "diamond.png"
        ).read_text()

    ############################################################################
    # low-level tests                                                          #
    ############################################################################

    def test_reading_the_target_of_a_remote_symlink(self, remote):
        address = remote._replace(
            path=urlparse(
                (
                    sync.abspath_from_uri(remote)
                    / "EnderChest"
                    / "optifine"
                    / "mods"
                    / "BME.jar"
                ).as_uri()
            ).path,
        )
        with sync.remote_file(address) as remote_link:
            assert remote_link.readlink().name == "BME_1.19.2_nightly.jar"

    @pytest.mark.parametrize("target_type", ("file", "symlink", "nothing"))
    def test_pull_replaces_existing_(self, target_type, remote, tmp_path):
        remote_chest = remote._replace(
            path=urlparse((sync.abspath_from_uri(remote) / "EnderChest").as_uri()).path,
        )
        local_path = tmp_path / "somewhere_else" / "EnderChest"
        local_path.parent.mkdir(parents=True)
        if target_type == "file":
            local_path.write_text("Leave me alone!\n")
        elif target_type == "symlink":
            # shouldn't need to exist
            local_path.symlink_to(tmp_path / "aether", target_is_directory=False)

        sync.pull(remote_chest, local_path.parent, verbosity=-1)
        assert (local_path / "enderchest.cfg").exists()

    @pytest.mark.parametrize("target_type", ("file", "symlink"))
    def test_pull_dry_run_does_not_replace_existing_(
        self, target_type, remote, tmp_path
    ):
        remote_chest = remote._replace(
            path=urlparse((sync.abspath_from_uri(remote) / "EnderChest").as_uri()).path,
        )
        local_path = tmp_path / "somewhere_else" / "EnderChest"
        local_path.parent.mkdir(parents=True)
        if target_type == "file":
            local_path.write_text("Leave me alone!\n")
        elif target_type == "symlink":
            # shouldn't need to exist
            local_path.symlink_to(tmp_path / "aether", target_is_directory=False)

        sync.pull(remote_chest, local_path.parent, verbosity=-1, dry_run=True)
        if target_type == "file":
            assert local_path.read_text("UTF-8") == "Leave me alone!\n"
        elif target_type == "symlink":
            assert local_path.readlink().name == "aether"
        else:
            assert not local_path.exists()

    def test_pull_fails_if_local_parent_folder_does_not_exist(self, remote):
        remote_chest = remote._replace(
            path=urlparse((sync.abspath_from_uri(remote) / "EnderChest").as_uri()).path,
        )
        local_path = Path("i do not exist")

        with pytest.raises(FileNotFoundError):
            sync.pull(remote_chest, local_path)

    def test_pull_fails_if_remote_does_not_exist(self, tmp_path):
        local_path = tmp_path / "i do not exist" / "wait yes i do"
        local_path.parent.mkdir(parents=True)
        local_path.touch()

        remote_path = Path("i do not exist")
        remote = ParseResult(
            scheme=self.protocol,
            netloc=sync.get_default_netloc(),
            path=urlparse(remote_path.absolute().as_uri()).path,
            params="",
            query="",
            fragment="",
        )
        with pytest.raises(FileNotFoundError):
            sync.pull(remote, local_path.parent.parent, verbosity=-1)

        assert local_path.exists()

    @pytest.mark.parametrize("target_type", ("file", "symlink", "nothing"))
    def test_push_replaces_existing_(self, target_type, minecraft_root, tmp_path):
        remote_path = tmp_path / "somewhere_else" / "EnderChest"
        remote_path.parent.mkdir(parents=True)
        if target_type == "file":
            remote_path.write_text("Leave me alone!\n")
        elif target_type == "symlink":
            # shouldn't need to exist
            remote_path.symlink_to(tmp_path / "aether", target_is_directory=False)

        remote = ParseResult(
            scheme=self.protocol,
            netloc=sync.get_default_netloc(),
            path=urlparse(remote_path.absolute().parent.as_uri()).path,
            params="",
            query="",
            fragment="",
        )

        sync.push(minecraft_root / "EnderChest", remote, verbosity=-1)

        assert (remote_path / "global" / "usercache.json").exists()

    @pytest.mark.parametrize("target_type", ("file", "symlink", "nothing"))
    def test_push_dry_run_does_not_replace_existing_(
        self, target_type, minecraft_root, tmp_path
    ):
        remote_path = tmp_path / "somewhere_else" / "EnderChest"
        remote_path.parent.mkdir(parents=True)
        if target_type == "file":
            remote_path.write_text("Leave me alone!\n")
        elif target_type == "symlink":
            # shouldn't need to exist
            remote_path.symlink_to(tmp_path / "aether", target_is_directory=False)

        remote = ParseResult(
            scheme=self.protocol,
            netloc=sync.get_default_netloc(),
            path=urlparse(remote_path.absolute().parent.as_uri()).path,
            params="",
            query="",
            fragment="",
        )

        sync.push(minecraft_root / "EnderChest", remote, verbosity=-1, dry_run=True)

        if target_type == "file":
            assert remote_path.read_text("utf-8") == "Leave me alone!\n"
        elif target_type == "symlink":
            assert remote_path.readlink().name == "aether"
        else:
            assert not remote_path.exists()

    def test_push_fails_if_local_does_not_exist(self, tmp_path):
        remote_path = tmp_path / "i do not exist" / "wait yes i do"
        remote_path.parent.mkdir(parents=True)
        remote_path.touch()

        local_path = Path("i do not exist")
        remote = ParseResult(
            scheme=self.protocol,
            netloc=sync.get_default_netloc(),
            path=urlparse(remote_path.absolute().parent.parent.as_uri()).path,
            params="",
            query="",
            fragment="",
        )
        with pytest.raises(FileNotFoundError):
            sync.push(local_path, remote, verbosity=-1)

        assert remote_path.exists()


class TestFileSyncOnly:
    def test_push_fails_if_remote_parent_folder_does_not_exist(self, minecraft_root):
        remote_path = Path("i do not exist").absolute()
        remote = urlparse(remote_path.as_uri())

        with pytest.raises(FileNotFoundError):
            sync.push(minecraft_root / "EnderChest", remote)


@pytest.mark.skipif(
    not shutil.which("rsync"), reason="rsync module cannot be imported on this system"
)
class TestRsyncVersionChecking:
    class FakeSubprocessResult:
        def __init__(self, stdout: str, stderr: str):
            self.stdout = stdout.encode("utf-8")
            self.stderr = stderr.encode("utf-8")

    def test_raises_runtime_error_if_rsync_is_not_installed(self, monkeypatch):
        from enderchest.sync import rsync

        original_run = subprocess.run

        def run_something_else(commands, **kwargs) -> subprocess.CompletedProcess:
            return original_run(["adsqwcvqawf"], **kwargs)

        monkeypatch.setattr(subprocess, "run", run_something_else)

        with pytest.raises(RuntimeError, match="is not installed"):
            _ = rsync._get_rsync_version()

    def test_raises_runtime_error_if_rsync_v_produces_error_message(
        self, monkeypatch
    ) -> None:
        from enderchest.sync import rsync

        def mock_run(
            commands, **kwargs
        ) -> TestRsyncVersionChecking.FakeSubprocessResult:
            return TestRsyncVersionChecking.FakeSubprocessResult(
                "irrelephant", "Something bad!!"
            )

        monkeypatch.setattr(subprocess, "run", mock_run)

        with pytest.raises(RuntimeError, match="Something bad!!"):
            _ = rsync._get_rsync_version()

    def test_raises_runtime_error_if_rsync_v_produces_no_output(
        self, monkeypatch
    ) -> None:
        from enderchest.sync import rsync

        def mock_run(
            commands, **kwargs
        ) -> TestRsyncVersionChecking.FakeSubprocessResult:
            return TestRsyncVersionChecking.FakeSubprocessResult("", "")

        monkeypatch.setattr(subprocess, "run", mock_run)

        with pytest.raises(RuntimeError, match="could not be executed"):
            _ = rsync._get_rsync_version()

    def test_raises_runtime_error_if_head_doesnt_match_regex(self, monkeypatch) -> None:
        from enderchest.sync import rsync

        def mock_run(
            commands, **kwargs
        ) -> TestRsyncVersionChecking.FakeSubprocessResult:
            return TestRsyncVersionChecking.FakeSubprocessResult(
                "rsync is totally installed on this system", ""
            )

        monkeypatch.setattr(subprocess, "run", mock_run)

        with pytest.raises(RuntimeError, match="Could not parse version output"):
            _ = rsync._get_rsync_version()

    def test_correctly_parses_version_string(self, monkeypatch) -> None:
        from enderchest.sync import rsync

        def mock_run(
            commands, **kwargs
        ) -> TestRsyncVersionChecking.FakeSubprocessResult:
            return TestRsyncVersionChecking.FakeSubprocessResult(
                "rsync  version 1.20.78163  protocol version whateva", ""
            )

        monkeypatch.setattr(subprocess, "run", mock_run)

        assert rsync._get_rsync_version() == (1, 20)


@pytest.mark.skipif(
    not shutil.which("rsync"), reason="rsync is not installed on this system"
)
class TestRsyncSync(TestFileSync):
    protocol = "rsync"

    def test_rsync_summary_summarizes_at_the_shulker_box_level(
        self, minecraft_root, remote, caplog
    ):
        caplog.set_level(logging.DEBUG)
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", dry_run=True)
        info_log = ""
        debug_log = ""
        for record in caplog.records:
            if record.levelname == "INFO":
                info_log += record.msg % record.args + "\n"
            elif record.levelname == "DEBUG":
                debug_log += record.msg % record.args + "\n"

        # meta-test--make sure that the creation is actually happening
        assert (
            "+++ " + os.path.sep.join(("EnderChest", "1.19", ".bobby", "chunk"))
            in debug_log
        )

        # meta-test--make sure nothing in the report at the shulker+1 level
        assert os.path.sep.join(("EnderChest", "1.19", ".bobby")) not in info_log

        assert (
            f"Within EnderChest{os.path.sep}1.19..."
            "\n  - Creating 1 file"
            "\n  - Updating 1 file"  # there's also the save symlink
            "\n  - Deleting 0 files"
        ) in info_log

    def test_rsync_summary_doesnt_report_details_if_the_whole_shulker_is_being_created(
        self, minecraft_root, remote, caplog
    ):
        caplog.set_level(logging.DEBUG)
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", dry_run=True)
        info_log = ""
        debug_log = ""
        for record in caplog.records:
            if record.levelname == "INFO":
                info_log += record.getMessage() + "\n"
            elif record.levelname == "DEBUG":
                debug_log += record.getMessage() + "\n"

        # meta-test--make sure that the creation is actually happening
        assert (
            "+++ "
            + os.path.sep.join(("EnderChest", "optifine", "mods", "optifine.jar"))
            in debug_log
        )

        # meta-test--make sure nothing in the report at the shulker+1 level
        assert os.path.sep.join(("EnderChest", "optifine", "mods")) not in info_log

        assert f"Creating EnderChest{os.path.sep}optifine" in info_log

    @pytest.mark.parametrize("verbosity", ("v", "vv", "vvv"))
    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_verbose_dry_run_doesnt_summarize(
        self, monkeypatch, minecraft_root, remote, caplog, op, verbosity
    ):
        def mock_summarize(*args, **kwargs):
            raise AssertionError("I was not to be called")

        from enderchest.sync import rsync

        monkeypatch.setattr(rsync, "summarize_rsync_report", mock_summarize)

        caplog.set_level(logging.DEBUG)
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, dry_run=True, verbosity=len(verbosity))

        debug_log = "\n".join(
            record.getMessage()
            for record in caplog.records
            if record.levelname == "DEBUG"
        )

        # this wouldn't be in the summary
        assert f"EnderChest{os.sep}global{os.sep}config" in debug_log

    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_quiet_dry_run_still_reports_stats(
        self, minecraft_root, remote, caplog, op
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, dry_run=True, verbosity=-1)

        printed_log = "\n".join(
            record.getMessage()
            for record in caplog.records
            if record.levelno > logging.INFO
        )

        assert "Number of created files" in printed_log

    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_super_quiet_dry_run_still_reports_stats(
        self, minecraft_root, remote, caplog, op
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, dry_run=True, verbosity=-1)

        printed_log = "\n".join(
            record.msg for record in caplog.records if record.levelno > logging.INFO
        )

        assert "Number of created files" in printed_log

    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_regular_sync_only_reports_overall_progress(
        self, minecraft_root, remote, capfd, op
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op)

        printed_log = capfd.readouterr().out

        assert f"EnderChest{os.sep}global" not in printed_log

        # but, like, make sure that it prints *something*
        assert "Number of created files" in printed_log

    @pytest.mark.parametrize("verbosity", ("v", "vv", "vvv"))
    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_verbose_sync_reports_file_level_progress(
        self, minecraft_root, remote, capfd, op, verbosity
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, verbosity=len(verbosity))

        printed_log = capfd.readouterr().out

        assert f"EnderChest{os.sep}global" in printed_log
        assert "xfr#2, to-chk=" in printed_log

    @pytest.mark.parametrize("quietude", ("q", "qq", "qqq"))
    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_quiet_sync_is_silent(self, minecraft_root, remote, capfd, op, quietude):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, verbosity=-len(quietude))

        printed_log = capfd.readouterr().out

        assert printed_log == ""


def _is_paramiko_installed() -> bool:
    try:
        import paramiko

        return True
    except ImportError:
        return False


PARAMIKO_INSTALLED = _is_paramiko_installed()


@pytest.mark.skipif(
    not PARAMIKO_INSTALLED, reason="EnderChest was not installed with SFTP support"
)
class TestSFTPSync(TestFileSync):
    protocol = "sftp"

    @pytest.fixture(autouse=True)
    def patch_paramiko(self, remote, monkeypatch, use_local_ssh):
        if not use_local_ssh:
            from enderchest.sync import sftp

            mock_sftp = mock_paramiko.MockSFTP(
                sync.abspath_from_uri(remote) / "EnderChest"
            )

            monkeypatch.setattr(
                sftp, "connect", mock_paramiko.generate_mock_connect(mock_sftp)
            )
            monkeypatch.setattr(sftp, "rglob", mock_paramiko.mock_rglob)

    @pytest.fixture(autouse=False)
    def generate_lstat_cache(self, remote):
        from enderchest.sync import sftp

        with sftp.connect(remote) as sftp_client:
            stats = sftp.get_contents(
                sftp_client, (sync.abspath_from_uri(remote) / "EnderChest").as_posix()
            )

        for path, sftp_attr in stats:
            sftp_attr.filename = path.as_posix()
        with as_file(LSTAT_CACHE) as cache_file:
            cache_file.write_text(
                json.dumps(
                    [
                        {
                            field: getattr(sftp_attr, field)
                            for field in mock_paramiko.CachedStat._fields
                        }
                        for path, sftp_attr in stats
                    ],
                    indent=4,
                )
            )

    # @pytest.mark.usefixtures("generate_lstat_cache")
    # def test_generate_lstat_cache(self):
    #     """Force cache gen, but just do it once"""
    #     pass

    def test_push_fails_if_remote_parent_folder_does_not_exist(self, minecraft_root):
        remote_path = Path("i do not exist").absolute()
        remote = ParseResult(
            scheme=self.protocol,
            netloc=sync.get_default_netloc(),
            path=urlparse(remote_path.absolute().as_uri()).path,
            params="",
            query="",
            fragment="",
        )
        with pytest.raises(FileNotFoundError):
            sync.push(minecraft_root / "EnderChest", remote)



"""Non-implementation-specific syncing utilities"""

import fnmatch
import getpass
import os
import socket
import stat
from collections import defaultdict
from collections.abc import Collection, Generator, Iterable
from enum import Enum, auto
from pathlib import Path
from typing import Any, Protocol, TypeVar
from urllib.parse import ParseResult, unquote
from urllib.request import url2pathname

from ..loggers import SYNC_LOGGER


def get_default_netloc() -> str:
    """Compile a netloc from environment variables, etc.

    Returns
    -------
    str
        The default netloc, which is {user}@{hostname}
    """
    return f"{getpass.getuser()}@{socket.gethostname()}".lower()


def abspath_from_uri(uri: ParseResult) -> Path:
    """Extract and unquote the path component of a URI to turn it into an
    unambiguous absolute `pathlib.Path`

    h/t https://stackoverflow.com/a/61922504

    Parameters
    ----------
    uri : ParseResult
        The parsed URI to extract the path from

    Returns
    -------
    Path
        The path part of the URI as a Path
    """
    host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=uri.netloc)
    return Path(os.path.abspath(os.path.join(host, url2pathname(unquote(uri.path)))))


def uri_to_ssh(uri: ParseResult) -> str:
    """Convert a URI to an SSH address

    Parameters
    ----------
    uri: ParseResult
        The URI to convert

    Returns
    -------
    str
        The SSH-format address
    """
    return "{user}{host}:{path}".format(
        user=f"{uri.username}@" if uri.username else "",
        host=(uri.hostname or "localhost") + (f":{uri.port}" if uri.port else ""),
        path=uri.path,
    )


def render_remote(alias: str, uri: ParseResult) -> str:
    """Render a remote to a descriptive string

    Parameters
    ----------
    alias : str
        The name of the remote
    uri : ParseResult
        The parsed URI for the remote

    Returns
    -------
    str
        `{uri_string} [({alias})]}`
            (alias is omitted if it's the same as the URI's hostname)
    """
    uri_string = uri.geturl()

    if uri.hostname != alias:
        uri_string += f" ({alias})"
    return uri_string


class _StatLike(Protocol):  # pragma: no cover
    @property
    def st_mode(self) -> int | None: ...

    @property
    def st_size(self) -> float | None: ...

    @property
    def st_mtime(self) -> float | None: ...


def is_identical(object_one: _StatLike, object_two: _StatLike) -> bool:
    """Determine if two objects are identical (meaning: skip when syncing)

    Parameters
    ----------
    object_one : os.stat_result or similar
        The first object to compare
    object_two : os.stat_result or similar
        The second object to compare

    Returns
    -------
    bool
        False if the objects are conclusively different, True otherwise.

    Notes
    -----
    As most implementations of the SFTP protocol do not include the check-file
    extension, this method is limited in what it can compare. Use with caution.
    """
    if stat.S_ISDIR(object_one.st_mode or 0) != stat.S_ISDIR(object_two.st_mode or 0):
        return False
    if stat.S_ISLNK(object_one.st_mode or 0) != stat.S_ISLNK(object_two.st_mode or 0):
        return False

    if stat.S_ISLNK(object_one.st_mode or 0):
        # there's no way from the stat to tell if two links have the same target
        # so err on the side of "nope"
        return False

    if stat.S_ISREG(object_one.st_mode or 0):
        # these comparisons should only be run on files
        if int(object_one.st_size or 0) != int(object_two.st_size or 0):
            return False
        if int(object_one.st_mtime or 0) != int(object_two.st_mtime or 0):
            return False
    return True


class Operation(Enum):
    """The recognized sync operations

    Notes
    -----
    There's no `UPDATE` operation because so far this class isn't used by
    anything that _can_ perform a delta update on a file
    """

    CREATE = auto()
    REPLACE = auto()
    DELETE = auto()


PathInfo = TypeVar(
    "PathInfo",
    tuple[Path, Any],
    tuple[str, Any],
    # TODO: the proper type hint is tuple[Path, *tuple[Any, ...]]
    #       but that's not supported until Python 3.11
)


def filter_contents(
    contents: Iterable[PathInfo],
    exclude: Collection[str],
    prefix: Path | str | None = None,
) -> Generator[PathInfo, None, None]:
    """Apply an exclusion filter to a list of files

    Parameters
    ----------
    contents : list of (Path, ...) tuples
        The contents to filter
    exclude : list of str
        The patterns to exclude
    prefix : Path, optional
        If the contents are iterating over a subdirectory, providing the directory
        as the `prefix` will allow filtering to be performed on the full path.

    Yields
    ------
    (Path, ...) tuples
        The elements of the provided list, omitting the ones
        to be excluded
    """
    for path_info in contents:
        if not any(
            (
                fnmatch.fnmatch(
                    os.path.normpath(
                        os.path.join(prefix or "", path_info[0], "placeholder")
                    ),
                    os.path.join("*", pattern, "*"),
                )
                for pattern in exclude
            )
        ):
            yield path_info


def diff(
    source_files: Iterable[tuple[Path, _StatLike]],
    destination_files: Iterable[tuple[Path, _StatLike]],
) -> Generator[tuple[Path, _StatLike, Operation], None, None]:
    """Compute the "diff" between the source and destination, enumerating
    all the operations that should be performed so that the destination
    matches the source

    Parameters
    ----------
    source_files : list of (Path, stat_result) tuples
        The files and file attributes at the source
    destination_files : list of (Path, stat_result) tuples
        The files and file attributes at the destination

    Returns
    -------
    Generator of (Path, stat_result, Operation) tuples
        The files, their attributes and the operations that should be performed on each file

    Notes
    -----
    - The order of paths returned will match the order provided by the `source_files`
      except for the deletions, which will all come at the end and will be sorted
      from longest to shortest path (so that individual files are marked for deletion
      before their parent folders).
    - The attributes of each path will correspond to the *source* attributes for
      creations and replacements and to the *destination* attributes for the deletions
    """
    destination_lookup: dict[Path, _StatLike] = dict(destination_files)
    for file, source_stat in source_files:
        if file not in destination_lookup:
            yield file, source_stat, Operation.CREATE
        else:
            destination_stat = destination_lookup.pop(file)
            if not is_identical(source_stat, destination_stat):
                yield file, source_stat, Operation.REPLACE
            # else: continue

    for file, destination_stat in sorted(
        destination_lookup.items(), key=lambda x: -len(str(x[0]))
    ):
        yield file, destination_stat, Operation.DELETE


def generate_sync_report(
    content_diff: Iterable[tuple[Path, _StatLike, Operation]], depth: int = 2
) -> None:
    """Compile a high-level summary of the outcome of the `diff` method
    and report it to the logging.INFO level

    Parameters
    ----------
    content_diff : list of (Path, Operation) tuples
        The files and the operations that are to be performed on each file, as
        generated by the `diff` method
    depth : int, optional
        How many directories to go down from the root to generate the summary.
        Default is 2 (just report on top-level files and folders within the
        source folder).

    Returns
    -------
    None
    """
    summary: dict[Path, dict[Operation, int] | Operation] = defaultdict(
        lambda: {Operation.CREATE: 0, Operation.REPLACE: 0, Operation.DELETE: 0}
    )

    for full_path, path_stat, operation in content_diff:
        try:
            path_key = full_path.parents[-depth]
        except IndexError:  # then this doesn't go in a subdirectory
            summary[full_path] = operation
            continue

        entry = summary[path_key]
        if isinstance(entry, Operation):
            # then this is described by the top-level op
            continue
        if operation == Operation.CREATE and stat.S_ISDIR(path_stat.st_mode or 0):
            # don't count folder creations
            continue

        entry[operation] += 1

    for path_key, report in sorted(summary.items()):
        if isinstance(report, Operation):
            # nice that these verbs follow the same pattern
            SYNC_LOGGER.info(f"{report.name[:-1].title()}ing %s", path_key)
        else:
            SYNC_LOGGER.info(
                "Within %s...\n%s",
                path_key,
                "\n".join(
                    f"  - {op.name[:-1].title()}ing {count} file{'' if count == 1 else 's'}"
                    for op, count in report.items()
                ),
            )



"""Functionality for onboarding and updating new installations and instances"""

import itertools
import json
import logging
import os
import re
from collections.abc import Iterable
from configparser import ConfigParser, ParsingError
from pathlib import Path
from typing import Any, TypedDict
from urllib.parse import ParseResult

from . import filesystem as fs
from .enderchest import EnderChest, create_ender_chest
from .instance import InstanceSpec, normalize_modloader, parse_version
from .inventory import load_ender_chest
from .loggers import GATHER_LOGGER
from .prompt import prompt
from .shulker_box import _matches_version


def gather_minecraft_instances(
    minecraft_root: Path, search_path: Path, official: bool | None
) -> list[InstanceSpec]:
    """Search the specified directory for Minecraft installations and return
    any that are can be found and parsed

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder). This will be used to
        construct relative paths.
    search_path : Path
        The path to search
    official : bool or None
        Whether we expect that the instances found in this location will be:
          - from the official launcher (official=True)
          - from a MultiMC-style launcher (official=False)
          - a mix / unsure (official=None)

    Returns
    -------
    list of InstanceSpec
        A list of parsed instances

    Notes
    -----
    - If a minecraft installation is found but cannot be parsed
      (or parsed as specified) this method will report that failure but then
      continue on.
    - As a corollary, if _no_ valid Minecraft installations can be found, this
      method will return an empty list.
    """
    try:
        ender_chest = load_ender_chest(minecraft_root)
    except FileNotFoundError:
        # because this method can be called during crafting
        ender_chest = EnderChest(minecraft_root)
    GATHER_LOGGER.debug("Searching for Minecraft folders inside %s", search_path)
    instances: list[InstanceSpec] = []
    for folder in fs.minecraft_folders(search_path):
        folder_path = folder.absolute()
        GATHER_LOGGER.debug("Found minecraft installation at %s", folder)
        if official is not False:
            try:
                instances.append(gather_metadata_for_official_instance(folder_path))
                GATHER_LOGGER.info(
                    "Gathered official Minecraft installation from %s", folder
                )
                _check_for_allowed_symlinks(ender_chest, instances[-1])
                continue
            except ValueError as not_official:
                GATHER_LOGGER.log(
                    logging.DEBUG if official is None else logging.WARNING,
                    ("%s is not an official instance:\n%s", folder, not_official),
                )
        if official is not True:
            try:
                instances.append(gather_metadata_for_mmc_instance(folder_path))
                GATHER_LOGGER.info(
                    "Gathered MMC-like Minecraft installation from %s", folder
                )
                _check_for_allowed_symlinks(ender_chest, instances[-1])
                continue
            except ValueError as not_mmc:
                GATHER_LOGGER.log(
                    logging.DEBUG if official is None else logging.WARNING,
                    "%s is not an MMC-like instance:\n%s",
                    folder,
                    not_mmc,
                )
        GATHER_LOGGER.warning(
            "%s does not appear to be a valid Minecraft instance", folder_path
        )
    for i, mc_instance in enumerate(instances):
        try:
            instances[i] = mc_instance._replace(
                root=mc_instance.root.relative_to(minecraft_root.resolve())
            )
        except ValueError:
            # TODO: if not Windows, try making relative to "~"
            pass  # instance isn't inside the minecraft root
    if not instances:
        GATHER_LOGGER.warning(
            "Could not find any Minecraft instances inside %s", search_path
        )
    return instances


def gather_metadata_for_official_instance(
    minecraft_folder: Path, name: str = "official"
) -> InstanceSpec:
    """Parse files to generate metadata for an official Minecraft installation

    Parameters
    ----------
    minecraft_folder : Path
        The path to the installation's .minecraft folder
    name : str, optional
        A name or alias to give to the instance. If None is provided, the
        default name is "official"

    Returns
    -------
    InstanceSpec
        The metadata for this instance

    Raises
    ------
    ValueError
        If this is not a valid official Minecraft installation

    Notes
    -----
    This method will always consider this instance to be vanilla, with no
    modloader. If a Forge or Fabric executable is installed inside this
    instance, the precise name of that version of that modded minecraft
    will be included in the version list.
    """
    launcher_profile_file = minecraft_folder / "launcher_profiles.json"
    try:
        with launcher_profile_file.open() as lp_json:
            launcher_profiles = json.load(lp_json)
        raw_versions: list[str] = [
            profile["lastVersionId"]
            for profile in launcher_profiles["profiles"].values()
        ]
    except FileNotFoundError as no_json:
        raise ValueError(f"Could not find {launcher_profile_file}") from no_json
    except json.JSONDecodeError as bad_json:
        raise ValueError(
            f"{launcher_profile_file} is corrupt and could not be parsed"
        ) from bad_json
    except KeyError as weird_json:
        raise ValueError(
            f"Could not parse metadata from {launcher_profile_file}"
        ) from weird_json

    version_manifest_file = minecraft_folder / "versions" / "version_manifest_v2.json"
    try:
        with version_manifest_file.open() as vm_json:
            version_lookup: dict[str, str] = json.load(vm_json)["latest"]
    except FileNotFoundError as no_json:
        raise ValueError(f"Could not find {version_manifest_file}") from no_json
    except json.JSONDecodeError as bad_json:
        raise ValueError(
            f"{version_manifest_file} is corrupt and could not be parsed"
        ) from bad_json
    except KeyError:
        GATHER_LOGGER.warning(
            "%s has no latest-version lookup."
            "\nPlease check the parsed metadata to ensure that it's accurate.",
            version_manifest_file,
        )
        version_lookup = {}

    versions: list[str] = []
    groups: list[str] = ["vanilla"]
    for version in raw_versions:
        if version.startswith("latest-"):
            mapped_version = version_lookup.get(version[len("latest-") :])
            if mapped_version is not None:
                versions.append(parse_version(mapped_version))
                groups.append(version)
                continue
        versions.append(parse_version(version))

    return InstanceSpec(name, minecraft_folder, tuple(versions), "", tuple(groups), ())


def gather_metadata_for_mmc_instance(
    minecraft_folder: Path, instgroups_file: Path | None = None
) -> InstanceSpec:
    """Parse files to generate metadata for a MultiMC-like instance

    Parameters
    ----------
    minecraft_folder : Path
        The path to the installation's .minecraft folder
    instgroups_file : Path
        The path to instgroups.json. If None is provided, this method will
        look for it two directories up from the minecraft folder

    Returns
    -------
    InstanceSpec
        The metadata for this instance

    Raises
    ------
    ValueError
        If this is not a valid MMC-like Minecraft instance

    Notes
    -----
    If this method is failing to find the appropriate files, you may want
    to try ensuring that minecraft_folder is an absolute path.
    """
    mmc_pack_file = minecraft_folder.parent / "mmc-pack.json"
    try:
        with mmc_pack_file.open() as mmc_json:
            components: list[dict] = json.load(mmc_json)["components"]

        version: str | None = None
        modloader: str | None = None

        for component in components:
            match component.get("uid"), component.get("cachedName", ""):
                case "net.minecraft", _:
                    version = parse_version(component["version"])
                case "net.fabricmc.fabric-loader", _:
                    modloader = "Fabric Loader"
                case "org.quiltmc.quilt-loader", _:
                    modloader = "Quilt Loader"
                case ("net.minecraftforge", _) | (_, "Forge"):
                    modloader = "Forge"
                case _, name if name.endswith("oader"):
                    modloader = name
                case _:
                    continue
            modloader = normalize_modloader(modloader)[0]
        if version is None:
            raise KeyError("Could not find a net.minecraft component")
    except FileNotFoundError as no_json:
        raise ValueError(f"Could not find {mmc_pack_file}") from no_json
    except json.JSONDecodeError as bad_json:
        raise ValueError(
            f"{mmc_pack_file} is corrupt and could not be parsed"
        ) from bad_json
    except KeyError as weird_json:
        raise ValueError(
            f"Could not parse metadata from {mmc_pack_file}"
        ) from weird_json

    name = minecraft_folder.parent.name

    instance_groups: list[str] = []

    if name == "":
        GATHER_LOGGER.warning(
            "Could not resolve the name of the parent folder"
            " and thus could not load tags."
        )
    else:
        instgroups_file = (
            instgroups_file or minecraft_folder.parent.parent / "instgroups.json"
        )

        try:
            with instgroups_file.open() as groups_json:
                groups: dict[str, dict] = json.load(groups_json)["groups"]
            for group, metadata in groups.items():
                # interestingly this comes from the folder name, not the actual name
                if name in metadata.get("instances", ()):
                    instance_groups.append(group)

        except FileNotFoundError:
            GATHER_LOGGER.warning(
                "Could not find %s and thus could not load tags", instgroups_file
            )
        except json.JSONDecodeError:
            GATHER_LOGGER.warning(
                "%s is corrupt and could not be parsed for tags", instgroups_file
            )
        except KeyError:
            GATHER_LOGGER.warning("Could not parse tags from %s", instgroups_file)

    instance_cfg = minecraft_folder.parent / "instance.cfg"

    try:
        parser = ConfigParser(allow_no_value=True, interpolation=None)
        parser.read_string("[instance]\n" + instance_cfg.read_text())
        name = parser["instance"]["name"]
    except FileNotFoundError:
        GATHER_LOGGER.warning(
            "Could not find %s and thus could not load the instance name", instance_cfg
        )
    except ParsingError:
        GATHER_LOGGER.warning(
            "is corrupt and could not be parsed the instance name", instance_cfg
        )
    except KeyError:
        GATHER_LOGGER.warning(f"Could not parse instance name from %s", instance_cfg)

    if name == "":
        raise ValueError("Could not determine the name of the instance.")

    return InstanceSpec(
        name,
        minecraft_folder,
        (version,),
        modloader or "",
        tuple(instance_groups),
        (),
    )


SERVER_JAR_PATTERNS: tuple[str, ...] = (
    r"^(minecraft_server).([^-]*).jar$",  # vanilla naming as per official docs
    # (not much we can do with server.jar)
    r"^(forge)-([0-9\.]*)-([0-9\.]*).*\.jar$",
    r"^(fabric)-server-mc.([^-]*)-loader.([0-9\.]*)-launcher.([0-9\.]*).jar$",
    r"^(paper)-([^-]*)-([0-9]*).jar$",
    r"^(purpur)-([^-]*)-([0-9]*).jar$",
    r"^(spigot)-([^-]*).jar$",
)


class _JarFileMeta(TypedDict):
    modloader: str
    minecraft_versions: tuple[str]


def _gather_metadata_from_jar_filename(jar_name: str) -> _JarFileMeta:
    """

    Parameters
    ----------
    jar_name : str
        The filename of the server jar

    Returns
    -------
    dict with two entries :
        modloader : str
             The (display) name of the modloader (vanilla corresponds to "")
        minecraft_versions : tuple of single str
            The minecraft version of the instance (tupled for `InstanceSpec`
            compatibility).

    Notes
    -----
    The filename may contain additional metadata (such as the modloader version).
    That metadata is ignored.

    Raises
    ------
    ValueError
        If the filename doesn't conform to any known patterns and thus
        metadata cannot be extracted).
    """
    for pattern in SERVER_JAR_PATTERNS:
        if pattern_match := re.match(pattern, jar_name):
            modloader, version, *_ = pattern_match.groups()
            break
    else:
        raise ValueError(f"Could not parse metadata from jar filename {jar_name}")
    return {
        "modloader": normalize_modloader(modloader)[0],
        "minecraft_versions": (version,),
    }


def gather_metadata_for_minecraft_server(
    server_home: Path,
    name: str | None = None,
    tags: Iterable[str] | None = None,
    server_jar: Path | None = None,
) -> InstanceSpec:
    """Parse files (or user input) to generate metadata for a minecraft server
    installation

    Parameters
    ----------
    server_home : Path
        The working directory of the Minecraft server
    name : str, optional
        A name or alias to give to the server. If None is provided, the user
        will be prompted to enter it.
    tags : list of str, optional
        The tags to assign to the server. If None are specified, the user will
        be prompted to enter them.
    server_jar : Path, optional
        The path to the server JAR file. If None is provided, this method will
        attempt to locate it within the `server_home`.

    Returns
    -------
    InstanceSpec
        The metadata for this instance

    Raises
    ------
    ValueError
        If this is not a valid Minecraft server installation or the requisite
        metadata could not be parsed

    Notes
    -----
    This method extracts metadata entirely from the filename of the server jar
    file. Custom-named jars or executables in non-standard locations will
    require their metadata be added manually.
    """
    instance_spec: dict[str, Any] = {"root": server_home, "groups_": ("server",)}
    if server_jar is not None:
        jars: Iterable[Path] = (server_jar,)
    else:
        jars = sorted(
            filter(
                lambda jar: not jar.is_relative_to(server_home / "mods"),
                itertools.chain(server_home.rglob("*.jar"), server_home.rglob("*.JAR")),
            ),
            key=lambda jar: (len(jar.parts), -len(str(jar))),
        )

    failed_parses: list[Path] = []
    for jar in jars:
        GATHER_LOGGER.debug("Attempting to extract server metadata from %s", jar)
        try:
            instance_spec.update(_gather_metadata_from_jar_filename(jar.name.lower()))
            break
        except ValueError as parse_fail:
            GATHER_LOGGER.debug(parse_fail)
            failed_parses.append(jar)
    else:
        GATHER_LOGGER.warning(
            "Could not parse server metadata from:\n%s",
            "\n".join((f"  - {jar}" for jar in failed_parses)),
        )
    if "modloader" not in instance_spec:
        instance_spec["modloader"] = normalize_modloader(
            prompt(
                "What modloader / type of server is this?"
                "\ne.g. Vanilla, Fabric, Forge, Paper, Spigot, PurPur..."
            )
            .lower()
            .strip()
        )[0]
        instance_spec["minecraft_versions"] = (
            prompt(
                "What version of Minecraft is this server?\ne.g.1.20.4, 23w13a_or_b..."
            )
            .lower()
            .strip(),
        )
    if name is None:
        name = prompt(
            "Enter a name / alias for this server", suggestion=server_home.name
        )
        if name == "":
            name = server_home.name
    instance_spec["name"] = name

    if tags is None:
        tags = prompt(
            "Enter any tags you'd like to use to label the server, separated by commas"
            '(it will be tagged as "server" automatically).'
        )
        if tags == "":
            tags = ()
        else:
            tags = (tag.lower().strip() for tag in tags.split(","))
    instance_spec["tags_"] = tuple(tags)

    return InstanceSpec(**instance_spec)


def update_ender_chest(
    minecraft_root: Path,
    search_paths: Iterable[str | Path] | None = None,
    instance_type: str | None = None,
    remotes: (
        Iterable[str | ParseResult | tuple[str, str] | tuple[ParseResult, str]] | None
    ) = None,
    **server_meta,
) -> None:
    """Orchestration method that coordinates the onboarding of new instances or
    EnderChest installations

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder).
    search_paths : list of Paths, optional
        The local search paths to look for Minecraft installations within.
        Be warned that this search is performed recursively.
    instance_type : str, optional
        Optionally specify the type of the Minecraft instances you expect to find.
        Options are:

          - from the official launcher (`instance_type="official"`)
          - from a MultiMC derivative (`instance_type="mmc"`)
          - server (in which case, each search path will be accepted verbatim
            as the server's home directory) (`instance_type="server"`)

        If `None` is specified, this method will search for both official and
        MMC-style instances (but not servers).
    remotes : list of URIs or (URI, str) tuples, optional
        Any remotes you wish to register to this instance. When a (URI, str) tuple
        is provided, the second value will be used as the name/alias of the remote.
        If there is already a remote specified with the given alias, this method will
        replace it.
    **server_meta
        Pass-through for metadata to pass through to any gathered servers (such
        as name or jar location)
    """
    try:
        ender_chest = load_ender_chest(minecraft_root)
    except (FileNotFoundError, ValueError) as bad_chest:
        GATHER_LOGGER.error(
            f"Could not load EnderChest from {minecraft_root}:\n  {bad_chest}"
        )
        return
    for search_path in search_paths or ():
        match instance_type:
            case "server":
                instance = gather_metadata_for_minecraft_server(
                    Path(search_path), **server_meta
                )
                _ = ender_chest.register_instance(instance)
                continue
            case "official":
                official: bool | None = True
            case "mmc":
                official = False
            case None:
                official = None
            case _:
                raise NotImplementedError(
                    f"{instance_type} instances are not currently supported."
                )
        for instance in gather_minecraft_instances(
            minecraft_root, Path(search_path), official=official
        ):
            _ = ender_chest.register_instance(instance)

    for remote in remotes or ():
        try:
            if isinstance(remote, (str, ParseResult)):
                ender_chest.register_remote(remote)
            else:
                ender_chest.register_remote(*remote)
        except ValueError as bad_remote:
            GATHER_LOGGER.warning(bad_remote)

    create_ender_chest(minecraft_root, ender_chest)


def _check_for_allowed_symlinks(
    ender_chest: EnderChest, instance: InstanceSpec
) -> None:
    """Check if the instance:
        - is 1.20+
        - has not already blanket-allowed symlinks into the EnderChest

    and if it hasn't, offer to update the allow-list now *but only if* the user
    hasn't already told EnderChest "shut up I know what I'm doing."

    Parameters
    ----------
    ender_chest : EnderChest
        This EnderChest
    instance : InstanceSpec
        The instance spec to check
    """
    if ender_chest.offer_to_update_symlink_allowlist is False:
        return

    if not any(
        _needs_symlink_allowlist(version) for version in instance.minecraft_versions
    ):
        return
    ender_chest_abspath = os.path.realpath(ender_chest.root)

    symlink_allowlist = instance.root / "allowed_symlinks.txt"

    try:
        allowlist_contents = symlink_allowlist.read_text()
        already_allowed = ender_chest_abspath in allowlist_contents.splitlines()
        allowlist_needs_newline = not allowlist_contents.endswith("\n")
    except FileNotFoundError:
        already_allowed = False
        allowlist_needs_newline = False

    if already_allowed:
        return

    GATHER_LOGGER.warning(
        """
Starting with Minecraft 1.20, Mojang by default no longer allows worlds
to load if they are or if they contain symbolic links.
Read more: https://help.minecraft.net/hc/en-us/articles/16165590199181"""
    )

    response = prompt(
        f"Would you like EnderChest to add {ender_chest_abspath} to {symlink_allowlist}?",
        "Y/n",
    )

    if response.lower() not in ("y", "yes", ""):
        return

    with symlink_allowlist.open("a") as allow_file:
        if allowlist_needs_newline:
            allow_file.write("\n")
        allow_file.write(ender_chest_abspath + "\n")

    GATHER_LOGGER.info(f"{symlink_allowlist} updated.")


def _needs_symlink_allowlist(version: str) -> bool:
    """Determine if a version needs `allowed_symlinks.txt` in order to link
    to EnderChest. Note that this is going a little broader than is strictly
    necessary.

    Parameters
    ----------
    version: str
        The version string to check against

    Returns
    -------
    bool
        Returns False if the Minecraft version predates the symlink ban. Returns
        True if it doesn't (or is marginal).

    Notes
    -----
    Have I mentioned that parsing Minecraft version strings is a pain in the
    toucans?
    """
    # first see if it follows basic semver
    if _matches_version(">1.19", parse_version(version.split("-")[0])):
        return True
    if _matches_version("1.20.0*", parse_version(version.split("-")[0])):
        return True
    # is it a snapshot?
    if match := re.match("^([1-2][0-9])w([0-9]{1,2})", version.lower()):
        year, week = match.groups()
        if int(year) > 23:
            return True
        if int(year) == 23 and int(week) > 18:
            return True

    return False



from pathlib import Path

from setuptools import setup

import versioneer

long_description = (Path(__file__).parent / "README.md").read_text()

setup(
    name="enderchest",
    python_requires=">=3.10",
    description="syncing and linking for all your Minecraft instances",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Gili "OpenBagTwo" Barlev',
    url="https://github.com/OpenBagTwo/EnderChest",
    packages=[
        "enderchest",
        "enderchest.sync",
        "enderchest.test",
        "enderchest.test.testing_files",
    ],
    package_data={
        "enderchest.test": [
            "testing_files/*.cfg",
            "testing_files/*.json",
            "testing_files/*.txt",
        ]
    },
    entry_points={
        "console_scripts": [
            "enderchest = enderchest.cli:main",
        ]
    },
    license="GPL v3",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=["semantic-version>=2.7", "pathvalidate>=2.5"],
    extras_require={
        "test": ["pytest>=7", "coverage>=7"],
        "sftp": ["paramiko>=3.3"],
    },
)



"""Tests of the EnderChest uninstallation procedure"""

import logging
import os

import pytest

from enderchest import craft
from enderchest import filesystem as fs
from enderchest import inventory, place, uninstall

from . import utils


class TestBreakEnderChest:

    @pytest.fixture(autouse=True)
    def patch_break(self, monkeypatch):
        """Patch the actual break method to prevent accidental breakages
        and to assert that the break method was never called"""

        def mock_break(*args, **kwargs):
            raise AssertionError("I should not have been called")

        monkeypatch.setattr(uninstall, "_break", mock_break)

    def test_cant_break_what_doesnt_exist(self, tmp_path, caplog):
        uninstall.break_ender_chest(tmp_path)
        error_log = [
            record for record in caplog.records if record.levelno == logging.ERROR
        ]
        assert error_log[-1].message == "Aborting."
        assert len(error_log) == 2
        assert error_log[0].message.startswith("Could not load EnderChest")

    def test_break_aborts_for_a_chest_with_no_instances(self, tmp_path, caplog):
        craft.craft_ender_chest(
            tmp_path,
            instance_search_paths=[],  # gotta give a kwarg so it doesn't go interactive
        )
        uninstall.break_ender_chest(tmp_path)
        errorish_log = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert errorish_log[-1].message == "Aborting."
        assert len(errorish_log) == 2
        assert errorish_log[0].levelno == logging.WARNING
        assert "no instances" in errorish_log[0].message.lower()

    def test_break_asks_you_if_youre_really_sure(
        self, minecraft_root, home, monkeypatch, capsys, caplog
    ):
        script_reader = utils.scripted_prompt([""])
        monkeypatch.setattr("builtins.input", script_reader)

        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False),
            *utils.TESTING_SHULKER_CONFIGS
        )

        uninstall.break_ender_chest(minecraft_root)

        _ = capsys.readouterr()  # suppress outputs
        errorish_log = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert errorish_log[-1].message == "Aborting."
        assert errorish_log[-2].levelno == logging.WARNING
        assert errorish_log[-2].message.startswith("Are you sure")

        assert len(errorish_log) == 2


class TestPartiallyBreakEnderChest:
    @pytest.fixture(autouse=True)
    def patch_break(self, monkeypatch):
        """Patch the actual break method to prevent accidental breakages
        and to assert that the break method was never called"""

        def mock_break(*args, **kwargs):
            raise AssertionError("I should not have been called")

        monkeypatch.setattr(uninstall, "_break", mock_break)

    def test_cant_break_what_doesnt_exist(self, tmp_path, caplog):
        with pytest.raises(FileNotFoundError, match="No valid EnderChest"):
            uninstall.break_instances(tmp_path, ("potato", "infinite"))

    def test_break_aborts_for_a_chest_with_no_instances(self, tmp_path, caplog):
        craft.craft_ender_chest(
            tmp_path,
            instance_search_paths=[],  # gotta give a kwarg so it doesn't go interactive
        )
        uninstall.break_instances(tmp_path, ("bee", "Drowned"))
        errorish_log = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert errorish_log[-1].message.splitlines()[-1] == "Aborting."
        assert "no valid instances" in errorish_log[-1].message.lower()

    def test_break_asks_you_if_youre_really_sure(
        self, minecraft_root, home, monkeypatch, capsys, caplog
    ):
        script_reader = utils.scripted_prompt([""])
        monkeypatch.setattr("builtins.input", script_reader)

        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False),
            *utils.TESTING_SHULKER_CONFIGS
        )

        uninstall.break_instances(minecraft_root, ("bee", "Drowned"))

        _ = capsys.readouterr()  # suppress outputs
        errorish_log = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert errorish_log[-1].message == "Aborting."

        assert errorish_log[-2].levelno == logging.WARNING
        assert errorish_log[-2].message.startswith("Are you sure")
        assert errorish_log[-2].args == ("  - bee\n  - Drowned",)

        assert len(errorish_log) == 2

    def test_break_skips_invalid_instances(
        self, minecraft_root, home, monkeypatch, capsys, caplog
    ):
        script_reader = utils.scripted_prompt([""])
        monkeypatch.setattr("builtins.input", script_reader)

        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False),
            *utils.TESTING_SHULKER_CONFIGS
        )

        uninstall.break_instances(minecraft_root, ("bee", "drowned"))

        _ = capsys.readouterr()  # suppress outputs
        errorish_log = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert len(errorish_log) == 3
        assert errorish_log[-3].levelno == logging.WARNING
        assert errorish_log[-3].args == ("drowned",)
        assert errorish_log[-3].message.lower().endswith("skipping.")

    def test_break_deregisters_instances_afterwards(
        self, minecraft_root, home, monkeypatch, capsys, caplog
    ):
        script_reader = utils.scripted_prompt(["Y"])
        monkeypatch.setattr("builtins.input", script_reader)

        def mock_break(*args, **kwargs):
            pass

        monkeypatch.setattr(uninstall, "_break", mock_break)

        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False),
            *utils.TESTING_SHULKER_CONFIGS
        )

        assert "bee" in (
            instance.name
            for instance in inventory.load_ender_chest_instances(minecraft_root)
        )

        uninstall.break_instances(minecraft_root, ("bee", "drowned"))

        _ = capsys.readouterr()  # suppress outputs

        assert "bee" not in (
            instance.name
            for instance in inventory.load_ender_chest_instances(minecraft_root)
        )


class TestBreaking:
    @pytest.fixture(autouse=True)
    def setup_teardown(self, minecraft_root, multi_box_setup_teardown, capsys, caplog):
        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False),
            *utils.TESTING_SHULKER_CONFIGS
        )
        place.place_ender_chest(minecraft_root, relative=True, error_handling="ignore")

        errorish_log = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert len(errorish_log) == 1
        assert (
            errorish_log[0].levelno == logging.ERROR
            and errorish_log[0].message.startswith("Error linking")
            and "options.txt" in errorish_log[0].message
        )

        yield

        _ = capsys.readouterr()  # suppress outputs

    def test_clean_uninstall_emits_no_warnings(self, minecraft_root, caplog):
        uninstall._break(fs.ender_chest_folder(minecraft_root), utils.TESTING_INSTANCES)
        warnings = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert len(warnings) == 0

    @pytest.mark.parametrize(
        "instance", (instance.name for instance in utils.TESTING_INSTANCES)
    )
    def test_uninstall_replaces_links_with_copies(self, minecraft_root, instance):
        resource_path = {
            instance.name: instance for instance in utils.TESTING_INSTANCES
        }[instance].root.expanduser() / "usercache.json"

        # meta-test
        assert resource_path.resolve().is_relative_to(
            fs.ender_chest_folder(minecraft_root.resolve())
        )
        uninstall._break(fs.ender_chest_folder(minecraft_root), utils.TESTING_INSTANCES)

        assert resource_path.exists()
        assert not resource_path.is_symlink()
        assert resource_path.read_text() == "alexander\nmomoa\nbateman\n"

    @pytest.mark.parametrize(
        "instance", (instance.name for instance in utils.TESTING_INSTANCES)
    )
    def test_uninstall_replaces_links_pointing_outside_the_chest_with_links(
        self, minecraft_root, instance
    ):
        resource_path = {
            instance.name: instance for instance in utils.TESTING_INSTANCES
        }[instance].root.expanduser() / "crash-reports"

        # fix for pytest utilizing symlinks for differently-scoped fixtures
        resource_path.unlink()
        resource_path.symlink_to(
            fs.ender_chest_folder(minecraft_root.resolve())
            / "global"
            / resource_path.name
        )

        # meta-test
        assert not resource_path.resolve().is_relative_to(
            fs.ender_chest_folder(minecraft_root.resolve())
        )

        uninstall._break(fs.ender_chest_folder(minecraft_root), utils.TESTING_INSTANCES)

        assert resource_path.exists()
        assert resource_path.is_symlink()
        assert (
            (resource_path / "20230524.log")
            .read_text()
            .endswith("And somehow also on fire\n")
        )

    @pytest.mark.parametrize("instance", ("axolotl", "bee"))
    def test_uninstall_copies_based_on_highest_priority(
        self,
        minecraft_root,
        instance,
    ):
        resource_path = (
            {instance.name: instance for instance in utils.TESTING_INSTANCES}[
                instance
            ].root.expanduser()
            / "resourcepacks"
            / "stuff.zip"
        )
        uninstall._break(fs.ender_chest_folder(minecraft_root), utils.TESTING_INSTANCES)

        assert resource_path.exists()
        assert resource_path.is_symlink() is (instance != "bee")
        assert (
            resource_path.read_text() == "optifine-optimized!"
            if instance == "bee"
            else "dfgwhgsadfhsd"
        )

    @pytest.mark.parametrize(
        "instance", (instance.name for instance in utils.TESTING_INSTANCES)
    )
    def test_uninstall_proceeds_past_broken_links(
        self, minecraft_root, instance, caplog
    ):
        (fs.shulker_box_root(minecraft_root, "global") / "backups").rmdir()

        resource_path = {
            instance.name: instance for instance in utils.TESTING_INSTANCES
        }["Chest Boat"].root.expanduser() / "logs"

        uninstall._break(fs.ender_chest_folder(minecraft_root), utils.TESTING_INSTANCES)

        errorish_log = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert len(errorish_log) == len(utils.TESTING_INSTANCES)

        assert all(
            (
                (
                    "No such file or directory" in record.message
                    or "system cannot find the file" in record.message
                )
                for record in errorish_log
            )
        )
        assert resource_path.exists()
        assert not resource_path.is_symlink()
        assert (
            (resource_path / "bumpona.log").read_text("utf-8").startswith("Like a bump")
        )

    def test_break_doesnt_touch_links_pointing_directly_outside_of_enderchest(
        self, minecraft_root, caplog
    ):
        # even broken ones
        resource_path = (
            {instance.name: instance for instance in utils.TESTING_INSTANCES}[
                "bee"
            ].root.expanduser()
            / "config"
            / "some_mod.json"
        )

        os.symlink(
            minecraft_root / "null" / "null", resource_path, target_is_directory=False
        )
        uninstall._break(fs.ender_chest_folder(minecraft_root), utils.TESTING_INSTANCES)
        warnings = [
            record for record in caplog.records if record.levelno >= logging.WARNING
        ]
        assert len(warnings) == 0

        assert resource_path.readlink().name == "null"



"""Tests for reading and writing config files"""

import warnings
from pathlib import Path
from urllib.parse import urlparse

import pytest

from enderchest import EnderChest, InstanceSpec, ShulkerBox
from enderchest import config as cfg
from enderchest import filesystem as fs
from enderchest.enderchest import _DEFAULTS
from enderchest.inventory import load_shulker_boxes
from enderchest.shulker_box import create_shulker_box
from enderchest.test import utils


class TestParseIniList:
    @pytest.mark.parametrize(
        "entry, expected",
        (
            ("hello", ["hello"]),
            ('"hello"', ["hello"]),
            ("'hello'", ["hello"]),
            ("  hello  ", ["hello"]),
            ("\nhello\n", ["hello"]),
            ("7", ["7"]),
            ("1,2,3", ["1", "2", "3"]),
            ("1, 2, 3,    ", ["1", "2", "3"]),
            ('"!", "\\"", ","', ["!", '"', ","]),
            ("yo\nmama\nis\n cool", ["yo", "mama", "is", "cool"]),
            ('yo\nmama\n  "  is cool  "', ["yo", "mama", "  is cool  "]),
            ("yo\nmama,papa\nis\ncool", ["yo", "mama,papa", "is", "cool"]),
        ),
        ids=(
            "single_str_uq",
            "single_str_dq",
            "single_str_sq",
            "padded_single_str",
            "newline_padded_single_str",
            "single_val",
            "one_line_list",
            "one_line_list_w_extras",
            "punctuation_list",
            "newlines",
            "newlines_with_quotes",
            "newlines_with_commas",
        ),
    )
    def test_parse_list_from_ini(self, entry, expected):
        assert cfg.parse_ini_list(entry) == expected


class TestConfigWriting:
    def test_shulker_box_config_roundtrip(self, minecraft_root):
        original_shulker = ShulkerBox(
            3,
            "original",
            minecraft_root / "EnderChest" / "original",
            match_criteria=(
                ("minecraft", (">=1.12,<2.0",)),
                ("modloader", ("*",)),
                ("tags", ("aether", "optifine")),
                ("instances", ("aether legacy", "Paradise Lost")),
            ),
            link_folders=("screenshots", "logs"),
            do_not_link=(),  # YEAH BABY! LINK THAT shulkerbox.cfg!!
        )

        utils.pre_populate_enderchest(minecraft_root / "EnderChest")

        create_shulker_box(
            minecraft_root, original_shulker, dict(_DEFAULTS)["shulker_box_folders"]
        )

        parsed_boxes = load_shulker_boxes(minecraft_root)

        assert parsed_boxes == [original_shulker]
        assert parsed_boxes[0].do_not_link == ()

    def test_ender_chest_config_roundtrip(self, tmpdir):
        (tmpdir / "EnderChest").mkdir()

        original_ender_chest = EnderChest(
            urlparse(Path(tmpdir).absolute().as_uri()),
            name="tester",
            remotes=[
                "irc://you@irl/home/upstairs",
                ("file:///lockbox", "undisclosed location"),
            ],
            instances=utils.TESTING_INSTANCES,
        )

        original_ender_chest.sync_confirm_wait = 27
        original_ender_chest.place_after_open = False
        original_ender_chest.offer_to_update_symlink_allowlist = False
        original_ender_chest.do_not_sync = ["EnderChest/enderchest.cfg", "*.local"]

        original_ender_chest.write_to_cfg(
            Path(tmpdir) / "EnderChest" / "enderchest.cfg"
        )

        parsed_ender_chest = EnderChest.from_cfg(
            Path(tmpdir) / "EnderChest" / "enderchest.cfg"
        )
        assert parsed_ender_chest.__dict__ == original_ender_chest.__dict__

    def test_ender_chest_self_corrects_its_config(self, tmp_path):
        (tmp_path / "EnderChest").mkdir()

        original_ender_chest = EnderChest(
            tmp_path,
            name="tester",
            instances=(
                InstanceSpec(
                    "puppet",
                    Path("..") / "instances" / "puppet" / ".minecraft",
                    ("1.20.1-pre1",),
                    "vanilla",
                    (),
                    ("your", "it"),
                ),
                InstanceSpec(
                    "puppet",
                    Path("..") / "instances" / "puppeteer" / ".minecraft",
                    ("1.20.2",),
                    "fabric",
                    (),
                    (),
                ),
            ),
        )
        original_ender_chest.do_not_sync = ["*.local"]

        original_ender_chest.write_to_cfg(
            fs.ender_chest_config(tmp_path, check_exists=False)
        )
        config_text = fs.ender_chest_config(tmp_path).read_text("utf-8").splitlines()
        config_text.remove("place-after-open = True")
        fs.ender_chest_config(tmp_path).write_text("\n".join(config_text))

        parsed_ender_chest = EnderChest.from_cfg(fs.ender_chest_config(tmp_path))

        assert {
            "do-not-sync": parsed_ender_chest.do_not_sync,
            "place-after-open": parsed_ender_chest.place_after_open,
            "modloader": parsed_ender_chest.instances[0].modloader,
            "dupe_instance_name": parsed_ender_chest.instances[1].name,
        } == {
            "do-not-sync": ["EnderChest/enderchest.cfg", "*.local"],
            "place-after-open": False,
            "modloader": "",
            "dupe_instance_name": "puppet.1",
        }

    def test_sync_confirm_wait_true_is_rendered_as_prompt(self):
        original_ender_chest = EnderChest(
            Path("ignoreme"),
            name="tester",
        )
        original_ender_chest.sync_confirm_wait = True
        config_text = original_ender_chest.write_to_cfg().splitlines()
        for line in config_text:
            if line.startswith("sync-confirmation-time"):
                assert line == "sync-confirmation-time = prompt"
                break
        else:
            raise KeyError("sync-confirm-wait line not found in config")


class TestConfigParsing:
    @pytest.mark.xfail(reason="Believe it or not, this can be turned into a URI")
    def test_raise_when_address_cannot_be_turned_into_a_uri(self, tmp_path):
        config_file = tmp_path / "enderchest.cfg"
        config_file.write_text(
            "[properties]"
            "\nname=tester\n"
            r"address=¯\_(ツ)_/¯"
            "\nsync-protocol=file"
            "\nplace-after-open=no"
        )
        with pytest.raises(ValueError, match="is not a valid URI"):
            _ = EnderChest.from_cfg(config_file)

    @pytest.mark.parametrize("value", ("prompt", "CONFIRM"))
    def test_sync_confirm_wait_prompt_is_read_as_true(self, tmp_path, value):
        config_file = tmp_path / "enderchest.cfg"
        config_file.write_text(
            "[properties]"
            "\nname=tester"
            "\nplace-after-open=no"
            f"\nsync-confirmation-time={value}"
        )
        assert EnderChest.from_cfg(config_file).sync_confirm_wait is True

    def test_raise_on_invalid_sync_confirm_wait(self, tmp_path):
        config_file = tmp_path / "enderchest.cfg"
        config_file.write_text(
            "[properties]"
            "\nname=tester"
            "\nplace-after-open=no"
            "\nsync-confirmation-time=sounds like a good idea"
        )
        with pytest.raises(ValueError, match="Invalid value for"):
            _ = EnderChest.from_cfg(config_file).sync_confirm_wait

    def test_raise_when_remote_has_no_alias(self, tmp_path):
        config_file = tmp_path / "enderchest.cfg"
        config_file.write_text(
            "[properties]"
            "\nname=tester"
            "\nplace-after-open=no"
            "\n\n[remotes]"
            "\n/some/where"
        )

        with pytest.raises(
            ValueError, match="All remotes must have an alias specified"
        ):
            _ = EnderChest.from_cfg(config_file)



"""Helpers for parsing and writing INI-format config files"""

import ast
import datetime as dt
from collections.abc import Iterable, Mapping, Sequence
from configparser import ConfigParser, ParsingError
from io import StringIO
from pathlib import Path
from typing import Any

from ._version import get_versions


def get_configurator() -> ConfigParser:
    """Generate a configuration parser capable of reading or writing INI files

    Returns
    -------
    ConfigParser
        The pre-configured configurator
    """
    configurator = ConfigParser(
        allow_no_value=True, inline_comment_prefixes=(";",), interpolation=None
    )
    configurator.optionxform = str  # type: ignore
    return configurator


def read_cfg(config_file: Path) -> ConfigParser:
    """Read in a configuration file

    Parameters
    ----------
    config_file : Path
        The path to the configuration file

    Returns
    -------
    ConfigParser
        The parsed configuration

    Raises
    ------
    ValueError
        If the config file at that location cannot be parsed
    FileNotFoundError
        If there is no config file at the specified location
    """
    configurator = get_configurator()
    try:
        assert configurator.read(config_file)
    except ParsingError as bad_cfg:
        raise ValueError(f"Could not parse {config_file}") from bad_cfg
    except AssertionError as not_read:
        raise FileNotFoundError(f"Could not open {config_file}") from not_read
    return configurator


def dumps(
    header: str | None,
    properties: dict[str, Any],
    **sections: Mapping[str, Any] | Sequence[str],
) -> str:
    """Serialize a configuration into an INI-formatted string

    Parameters
    ----------
    header: str
        A header to render as a comment at the top of the file
    properties: dict
        The "main" section contents. Note that this method will add some of its own
    **sections : dict or list
        Any additional sections to write. Each section may consist of a set
        of key-value pairs or they might simply be a list of values

    Returns
    -------
    str
        The contents of the configuration, suitable for writing to file
    """
    config = get_configurator()

    config.add_section("properties")
    for key, value in properties.items():
        config.set("properties", to_ini_key(key), to_ini_value(value))

    config.set(
        "properties",
        "last-modified",
        to_ini_value(dt.datetime.now()),
    )
    config.set(
        "properties", "generated-by-enderchest-version", get_versions()["version"]
    )

    for section, values in sections.items():
        section_name = to_ini_key(section)
        config.add_section(section_name)
        if isinstance(values, Mapping):
            for key, value in values.items():
                config.set(section_name, to_ini_key(key), to_ini_value(value))
        else:
            for value in values:
                config.set(section_name, to_ini_value(value))

    buffer = StringIO()
    if header:
        buffer.write(f"; {header}\n")
    config.write(buffer)
    buffer.seek(0)  # rewind
    return buffer.read()


def to_ini_key(key: str) -> str:
    """Style guide enforcement for INI keys

    Parameters
    ----------
    key : str
        The entry key to normalize

    Returns
    -------
    str
        The normalized key
    """
    return key.replace("_", "-")


def to_ini_value(value: Any) -> str:
    """Format a value into a string suitable for use in an INI entry

    Parameters
    ----------
    value
        The value to format

    Returns
    -------
    str
        The formatted INI-appropriate value
    """
    if isinstance(value, str):
        # have to put in this check since strings are iterable
        return value
    if value is None:
        return ""
    if isinstance(value, Iterable):
        return list_to_ini(list(value))
    if isinstance(value, dt.datetime):
        return value.isoformat(sep=" ")
    if isinstance(value, dt.date):
        # note that datetimes are considered dates
        return value.isoformat()

    return str(value)


def list_to_ini(values: Sequence) -> str:
    """Format a list of values into a string suitable for use in an INI entry

    Parameters
    ----------
    values : list-like
        the values in the list

    Returns
    -------
    str
        The formatted INI-appropriate value
    """
    if len(values) == 0:
        return ""
    if len(values) == 1:
        return values[0]
    return "\n" + "\n".join(values)


def parse_ini_list(entry: str) -> list[str]:
    """Parse a list from an INI config entry

    Parameters
    ----------
    entry : str
        The raw entry from the INI

    Returns
    -------
    list of str
        The parsed entries

    Notes
    -----
    This method is *only* for parsing specific values of a key-value entry
    *and not* for the whole "section is the key, lines are the values" thing
    I've got going on.
    """
    entry = entry.strip()
    try:
        parsed = ast.literal_eval(entry)
        if isinstance(parsed, str):
            return [parsed]
        return [str(value) for value in parsed]
    except (TypeError, ValueError, SyntaxError):
        # if only it were that easy...
        pass

    values: list[str] = []
    for line in entry.splitlines():
        try:
            values.append(str(ast.literal_eval(line)))
        except (TypeError, ValueError, SyntaxError):
            values.append(line.strip())
    return values
